import 'dart:async';
import 'dart:convert';
import 'package:flutter/services.dart';
import 'package:ln_emitter/emitter.dart';

import 'codec.dart';

/// BridgeResponse
class BridgeResponse<T> {
  /// code
  final int code;

  /// message
  final String message;

  /// data
  final T data;

  BridgeResponse({
    this.code,
    this.message,
    this.data,
  });
}

/// BridgeError
class BridgeError extends Error {
  /// Bridge response.
  final dynamic response;

  /// Bridge method name.
  final String method;

  /// Bridge call params.
  final dynamic params;

  /// Error message.
  final String err;

  /// Invalid bridge response format error.
  static final String invalidResponseError = 'invalid response';

  /// Api error.
  static final String apiError = 'api error';

  BridgeError({
    this.err,
    this.method,
    this.params,
    this.response,
  });

  String toString() {
    // 错误日志会以第一行日志作为 identifier，所以把差异化详细详细写到第二行
    return 'BridgeError: method: $method, message: $err\r\n' +
        'params: ${jsonEncode(params)}\r\n' +
        'response: ${jsonEncode(response)}\r\n';
  }
}

/// IBridgeClient
abstract class IBridgeClient extends EventEmitter {
  Future<BridgeResponse<T>> call<T>(
    String method, {
    dynamic params,

    /// 用于兼容旧版 bridge 格式
    int successCode,
  });

  Future<T> getData<T>(
    String method, {
    dynamic params,

    /// 用于兼容旧版 bridge 格式
    int successCode,
  });

  String prefix() => 'TTR';
}

/// BridgeClient.
class BridgeClient extends EventEmitter implements IBridgeClient {
  /// Bridge channel.
  final MethodChannel _channel;

  /// Default channel name.
  static const String defaultChannelName =
      'com.bytedance.hybrid.bridge-flutter';

  /// Error emitter.
  final StreamController _errorEmitter =
      new StreamController.broadcast(sync: false);

  /// 是否运行抛异常，全局开关，如果关闭，业务方不会收到异常
  bool enableThrowException = false;

  // Mutable 兼容旧版
  int globalSuccessCode = 1;
  int globalErrorCode = 0;

  BridgeClient({
    String channelName = defaultChannelName,
  }) : _channel = MethodChannel(channelName, ComposeMethodCodec()) {
    _channel.setMethodCallHandler(_onEvent);
  }

  /// Channel handler.
  Future<dynamic> _onEvent(MethodCall call) {
    emit(call.method, call.arguments);
    Completer completer = Completer();
    completer.complete();
    return completer.future;
  }

  /// Emit error.
  void emitError(BridgeError err, {Exception ex}) {
    _errorEmitter.add(err);
    throw err;
  }

  /// Emit value
  BridgeResponse<T> emitValue<T>(
      String method, dynamic params, int successCode, dynamic resp) {
    if (resp is! Map) {
      return  BridgeResponse<T>(
        code: 1,
        message: 'success',
        data: resp,
      );
    }
    Map respMap = Map<String, dynamic>.from(resp);
    // 兼容旧版 bridge
    List<List<String>> compatibleKeys = [
      ['err_msg', 'message'],
      ['msg', 'message'],
      ['err_no', 'code'],
    ];
    compatibleKeys.forEach((pair) {
      if (respMap.containsKey(pair[0]) && !respMap.containsKey(pair[1])) {
        respMap[pair[1]] = respMap[pair[0]];
      }
    });

    if (!respMap.containsKey('code') ||
        !(respMap['code'] is int) ||
        (respMap.containsKey('message') && !(respMap['message'] is String))) {
      emitError(BridgeError(
        err: BridgeError.invalidResponseError,
        method: method,
        params: params,
        response: resp,
      ));
    }
    if (successCode == null) {
      successCode = globalSuccessCode;
    }
    final bridgeResp = BridgeResponse<T>(
      code: respMap['code'],
      message: respMap['message'] ?? '',
      data: respMap['data'],
    );
    if (bridgeResp.code != successCode) {
      emitError(BridgeError(
        err: BridgeError.apiError,
        method: method,
        params: params,
        response: bridgeResp,
      ));
    }

    return bridgeResp;
  }

  /// Bridge call.
  Future<BridgeResponse<T>> call<T>(
    String method, {
    dynamic params,

    /// 用于兼容旧版 bridge 格式
    int successCode,
  }) async {
    dynamic resp;

    try {
      resp = await _channel.invokeMethod(method, params);
    } catch (ex) {
      emitError(
        BridgeError(
          err: ex.toString(),
          method: method,
          params: params,
          response: resp,
        ),
        ex: ex,
      );
    }

    if (resp == null /* || !(resp is Map) */) {
      emitError(BridgeError(
        err: BridgeError.invalidResponseError,
        method: method,
        params: params,
        response: resp,
      ));
    }

    return emitValue(method, params, successCode, resp);
  }

  @override
  Future<T> getData<T>(String method, {dynamic params, int successCode}) async {
    final resp = await call(method, params: params, successCode: successCode);
    return resp?.data;
  }

  /// 监听全局异常信息
  void onError(
    void onData(dynamic event), {
    Function onError,
    void onDone(),
    bool cancelOnError,
  }) {
    _errorEmitter.stream.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  @override
  String prefix() => 'TTR';
}

/// Bridge Module.
class BridgeModule {
  final IBridgeClient client;

  BridgeModule({
    this.client,
  });
}
