import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';

/// [MethodCodec] with UTF-8 encoded JSON method calls and result envelopes.
///
/// Values supported as method arguments and result payloads are those supported
/// by [JSONMessageCodec].
class ComposeMethodCodec implements MethodCodec {
  static const int JSON_TYPE = 0;
  static const int NORMAL_TYPE = 1;

  // The codec serializes method calls, and result envelopes as outlined below.
  // This format must match the Android and iOS counterparts.
  //
  // * Individual values are serialized as defined by the JSON codec of the
  //   dart:convert package.
  // * Method calls are serialized as two-element maps, with the method name
  //   keyed by 'method' and the arguments keyed by 'args'.
  // * Reply envelopes are serialized as either:
  //   * one-element lists containing the successful result as its single
  //     element, or
  //   * three-element lists containing, in order, an error code String, an
  //     error message String, and an error details value.

  /// Creates a [MethodCodec] with UTF-8 encoded JSON method calls and result
  /// envelopes.
  const ComposeMethodCodec();

  @override
  ByteData encodeMethodCall(MethodCall call) {
    if (call.arguments != null && call.arguments['__is__standard__'] == true) {
      final codec = StandardMessageCodec();
      final WriteBuffer buffer = WriteBuffer();
      buffer.putUint8(NORMAL_TYPE);
      codec.writeValue(buffer, <String, dynamic>{
        'method': call.method,
        'args': call.arguments,
      });
      return buffer.done();
    }

    final origin = JSONMessageCodec().encodeMessage(<String, dynamic>{
      'method': call.method,
      'args': call.arguments,
    });
    final WriteBuffer buffer = WriteBuffer();
    buffer.putUint8(JSON_TYPE);
    buffer.putUint8List(
      origin.buffer.asUint8List(origin.offsetInBytes, origin.lengthInBytes),
    );
    return buffer.done();
  }

  @override
  MethodCall decodeMethodCall(ByteData methodCall) {
    final ReadBuffer buffer = ReadBuffer(methodCall);
    dynamic decoded;
    if (buffer.getUint8() == NORMAL_TYPE) {
      decoded = StandardMessageCodec().readValue(buffer);
    } else {
      decoded = const JSONMessageCodec().decodeMessage(methodCall);
    }

    if (decoded is! Map)
      throw FormatException('Expected method call Map, got $decoded');
    final dynamic method = decoded['method'];
    final dynamic arguments = decoded['args'];
    if (method is String) return MethodCall(method, arguments);
    throw FormatException('Invalid method call: $decoded');
  }

  /// native => flutter
  @override
  dynamic decodeEnvelope(ByteData envelope) {
    final ReadBuffer buffer = ReadBuffer(envelope);
    if (buffer.getUint8() == NORMAL_TYPE) {
      return StandardMessageCodec().readValue(buffer);
    }

    final dynamic decoded = const JSONMessageCodec().decodeMessage(envelope);
    if (decoded is! List)
      throw FormatException('Expected envelope List, got $decoded');
    if (decoded.length == 1) return decoded[0];
    if (decoded.length == 3 &&
        decoded[0] is String &&
        (decoded[1] == null || decoded[1] is String))
      throw PlatformException(
        code: decoded[0],
        message: decoded[1],
        details: decoded[2],
      );
    throw FormatException('Invalid envelope: $decoded');
  }

  bool isPrimaryType(dynamic result) {
    return result is bool ||
        result is String ||
        result is Uint8List ||
        result is Int8List ||
        result is Int16List ||
        result is Int32List ||
        result is Int64List ||
        result is Float64List ||
        result is List ||
        result is Map;
  }

  /// flutter => native.
  @override
  ByteData encodeSuccessEnvelope(dynamic result) {
    if (isPrimaryType(result)) {
      final codec = StandardMessageCodec();
      final WriteBuffer buffer = WriteBuffer();
      buffer.putUint8(NORMAL_TYPE);
      codec.writeValue(buffer, result);
      return buffer.done();
    }
    final origin = JSONMessageCodec().encodeMessage(<dynamic>[result]);
    final WriteBuffer buffer = WriteBuffer();
    buffer.putUint8(JSON_TYPE);
    buffer.putUint8List(
      origin.buffer.asUint8List(origin.offsetInBytes, origin.lengthInBytes),
    );
    return buffer.done();
  }

  @override
  ByteData encodeErrorEnvelope(
      {@required String code, String message, dynamic details}) {
    assert(code != null);
    return const JSONMessageCodec()
        .encodeMessage(<dynamic>[code, message, details]);
  }
}
