import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';

import './annotation.dart';

class BridgeGenerator extends GeneratorForAnnotation<Bridge> {
  @override
  String generateForAnnotatedElement(
    Element el,
    ConstantReader annotation,
    BuildStep step,
  ) {
    if (el is! ClassElement) {
      throw new AssertionError('@Bridge can only be decorated on classes.');
    }

    final ClassElement element = el;
    final className = el.displayName + 'Bridge';
    final bridgeName =
        annotation.read('value').literalValue as String ?? el.displayName;

    var lines = <String>[
      'class $className extends BridgeModule with ${el.displayName} {',
      '$className({ BridgeClient client }): super(client: client);',
    ];

    element.methods.forEach((method) {
      final metadata = method.metadata;

      metadata.forEach((e) => e.computeConstantValue());

      final exists = metadata.indexWhere(
              (el) => el.constantValue.type.name == 'BridgeMethod') ==
          -1;
      if (exists) {
        // lines.add(method.computeNode().toSource());
        return;
      }

      final annotation = metadata.firstWhere(
        (el) => el.constantValue.type.name == 'BridgeMethod',
      );
      final field = annotation.constantValue.getField('name');
      final methodName = field?.toStringValue() ?? method.displayName;
      final callRaw =
          annotation.constantValue.getField('callRaw')?.toBoolValue() ?? false;
      final callStandard =
          annotation.constantValue.getField('callStandard')?.toBoolValue() ?? false;
      final hasParams = method.parameters.isNotEmpty;

      var typeParameters = '';
      if (method.typeParameters.length > 0) {
        typeParameters =
            '<${method.typeParameters.map((el) => el.displayName).join(',')}>';
      }

      String returnValueType = "";
      if (method.returnType.isDartAsyncFuture) {
        returnValueType =
            method.returnType.displayName.replaceFirst('Future', '');
      }

      lines.add(
        '${method.returnType.displayName} ${method.displayName}$typeParameters(${hasParams ? '{' : '){'}',
      );

      if (hasParams) {
        method.parameters.forEach((param) {
          lines.add(
            '${param.hasRequired ? '@required ' : ''}${param.type.displayName} ${param.displayName},',
          );
        });

        lines.add('}) {');
        lines.add('var _\$params = {');

        method.parameters.forEach((param) {
          lines.add("'${param.displayName}': ${param.displayName},");
        });

        if (callStandard) {
          lines.add("'__is__standard__': true,");
        }

        lines.add('};');
      }

      // param processor
      final paramProcessor =
          annotation.constantValue.getField('paramProcessor')?.toStringValue();
      // result processor
      final resultProcessor =
          annotation.constantValue.getField('resultProcessor')?.toStringValue();

      if (resultProcessor != null) {
        final resultMethod = element.getMethod(resultProcessor);
        if (resultMethod.parameters.length == 1 && resultMethod.parameters[0].type.isDartAsyncFuture) {
          final inputType = resultMethod.parameters[0].type;
          returnValueType = inputType.displayName.replaceFirst('Future', '');
        }
      }

      if (paramProcessor != null && hasParams) {
        lines.add('_\$params = this.$paramProcessor(_\$params);');
      }
      lines.add(
        "var _\$result = client.${callRaw ? 'call' : 'getData'}$returnValueType('\$\{client.prefix()\}$bridgeName.$methodName'${hasParams ? ', params: _\$params' : ''});",
      );
      if (resultProcessor != null) {
        lines.add("var _\$resultFixed = this.$resultProcessor(_\$result);");
        lines.add('return _\$resultFixed;');
      } else {
        lines.add('return _\$result;');
      }
      lines.add('}');
    });
    lines.add('}');
    return lines.join('\n');
  }
}
